# -*- coding: utf-8 -*-
"""
Created on Thu Aug 18 13:26:56 2016

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import super
"""
from future import standard_library
standard_library.install_aliases()
"""
from .form import form
from ..transform.graphTransform import graphTransform, graphTransform3D

"""
 * A singly-form denotes a form of a singly-associated sort.
 * Singly-forms may contain only one individual.
 * Different types of singly-forms are distinguished based on the operational
 * behavior of their forms. 
 *
 * An {OrdinalForm} specifies an ordinal behavior on numbers: any set of 
 * numbers is reduced to the single maximum number, all other numbers in the 
 * original set are a part of this maximum number.
 *
 * An singly-form is maximal if it is not equal to nil and its
 * individual's attribute form is maximal. If it is equal to nil, it
 * should not have an individual, i.e., its individual should be null.
 * 
 * The SinglyForm class extends on the {Form} class.
 * It implements the form container in a trivial way, containing a single
 * individual.
 * It provides common implementations for some interface methods.
"""
class singlyForm(form):
    """
    initialize the form
    self.individual is the single individual that belongs to this form
    """
    def __init__(self, sort, *args, **kwargs):
        super(singlyForm, self).__init__(sort)
        self.individual = None
        self.maximal = True
        
    def __iter__(self):
        return iter([self.individual])
        
    def splitElements(self):
        return [self]
                
    def totalLength(self):
        if self.isEmpty():
            return 0
        return 1
        
    """
    check if the form is empty
    """
    def isEmpty(self):
        if self.individual is None:
            return True
        return self.individual.isNil()
        
    def __len__(self):
        if self.individual is None:
            return 0
        return 1
        
    def __str__(self):
        s = '{'
        if self.individual.sort.type == 'sort-attribute' and self.individual.hasAttribute():
            s += self.individual.toString(False) + str(self.individual.attribute)
        else:
            s += self.individual.toString(False)
        s += '}'
        return s
        
    def toString(self, string = ''):
        return str(self)
        
    def oneLineString(self, printKey = False):
        return str(self)
        
    def getElements(self):
        return [self.individual]
        
    """
    return the first element of the form which is the only individual
    """
    def first(self):
        return self.individual

    """
    return the last element of the form which is also the only individual
    """
    def last(self):
        return self.individual

    def next(self, current):
        return None

    def previous(self, current):
        return None

    """
    maximalize the form
    if the individual is Nil(empty, or has value of 0/min): remove the individual
    if the individual has attribute, maximalize the attribute form
    """
    def maximalize(self, ignoreMaximal = False):
        if self.maximal and ignoreMaximal == False:
            return
        self.maximal = True
        if self.individual is None:
            return
        if self.individual.isNil():
            self.individual = None
        elif self.individual.hasAttribute():
            self.individual.attribute.maximalize()
    """
    check if two forms are equal
    if the individual in the two forms are equal, the two forms are equal 
    """
    def __eq__(self, other):
        if other is None or other.__class__ != self.__class__ or not other.sort == self.sort:
            return False
        if self.individual is None:
            if other.individual is None:
                return True
            return False
        if other.individual is None:
            return False
        if not self.maximal:
            self.maximalize()
        if not other.maximal:
            other.maximalize()
        if self.individual == other.individual:
            if self.sort.type == 'sort-attribute':
                return self.individual.attribute == other.individual.attribute
            return True
        return False
        
    """
    merge the other form into this form
    i.e. changing this form's individual into the other's individual if the 
         other's individua is greater than this form's individual
    """
    def merge(self, other):
        if other.individual.compareValue(self.individual) == 1:
            self.add(other.individual)
        return
        
    """
    make the individual into this form's individual
    """        
    def add(self, individual):
        self.maximal = False
        if individual.assoc is None:
            self.individual = individual
            individual.assoc = self
        else:
            newCopy = individual.duplicateInd() 
            self.individual = newCopy
            newCopy.assoc = self
        
    """
    remove the individual from this form (make it minimum)
    """
    def remove(self):
        try:
            self.individual = self.individual.getMinInd()        
        except AttributeError:
            self.individual = None
    
    """
    check this form's individual for rule analysis or rule detection
    """
    def accept(self, visitor, direction=-1):
        if not self.isEmpty() and visitor.visitEnter(self, direction):
            self.maximalize()
            self.individual.accept(visitor, direction)
        visitor.visitLeave(self)
        return
        
    """
    editing the form in a certain procedure(editor)
    mainly use for transforming the form elements with transformation matrix(es)
    """        
    def edit(self, editor):
        result = self.sort.newForm()
        result.individual = self.individual.edit(editor)
        if editor.current is None:
            transf = editor.retrieveTransform(graphTransform)
            if transf is not None:
                for i in transf.extra:
                    result.insert(i)
            transf = editor.retrieveTransform(graphTransform3D)
            if transf is not None:
                for i in transf.extra:
                    result.insert(i)
        return result
        
    """
    converting all individuals inside this form that uses vectors with coordinates into 
    using vectors with float values 
    """
    def convertFloat(self):
        result = self.sort.newForm()
        result.individual = self.individual.convertFloat()
        return result
        
    """
    check when this form's individual is converted to float value, if it has 
    the same value as the element to be found
    """
    def findFloatValue(self, element):
        if self.sort == element.sort:
            return self.individual.findFloatValue(element)
        return False
        
    """
    check if this form contains a certain individual
    i.e. this form's individual is ind
    """
    def containsInd(self, ind):
        if self.individual == ind:
            return [self.individual]
        return False
        
    """
    check if this form is contained by the compareForm after transforming
    with transformSet
    """
    def checkTransform(self, transformSet, compareForm, flagDesc = True):
        if self.individual.checkTransform(transformSet, compareForm):
            return True
        return False           
        
    """
    clear the content of the form
    """
    def clear(self):
        if self.individual is not None:
            self.individual.assoc = None
            self.individual = None
        
    def duplicateForm(self):
        newForm = self.sort.newForm()
        newForm.insert(self.individual.duplicate())
        return newForm
    #------------------------------------------------------------------
    
"""    
    def hashcode(self):
        return
    def toString(self):
        return
    def parse(self):
        return 
    def clear(self):
        return 
"""

"""
    def insert(self):
        return
    def merge(self):
        return 
""" 
        