# -*- coding: utf-8 -*-
"""
Created on Tue Aug 23 12:23:02 2016

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import super
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from .line import line2D, line3D
from ..struct.coordinate import c0p5, c2
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..io.parseReader import ParsingError

"""
 * A line-segment is a connected and bounded segment of a line with an
 * interval behavior. The line defines the co-descriptor of the line-segment,
 * the boundary of the segment is defined by the start and end positions of the
 * line-segment.
 * 
 * The LineSegment class extends on the Line class and
 * implements the Interval interface.
 * It defines the characteristic individual for line-segments.
 * A line-segment is represented as a line with two rational
 * scalars specifying the tail and head relative to the line's root.
 * This characteristic individual accepts no parameters.
 * It specifies a {sortal.map.similarityMap} as default.
 * Forms of line-segments adhere to an interval behavior.
"""

class lineSegment2D(line2D):
    """
    initialize a line segment
    initialize a line containing it first
    find the lesser and greater points between the 2 points passed in (start 
    and end) and store them as self.start and self.end respectively
    """
    type = 'ind-lineSegment2D'
    
    def __init__(self, sort, start, end, *args, **kwargs):
        if end.type == 'ind-point2D':
            end = end.value
        if start.type == 'ind-point2D':
            start = start.value        
        if end.compareValue(start) == -1:
            self.start = end
            self.end = start
            self.reverse = True
        elif end.compareValue(start) == 1:
            self.end = end
            self.start = start
            self.reverse = False
        else:
            #print('___',start.printRhino(), end.printRhino())
            raise ValueError('start point and end point of the line segment are the same')
        super(lineSegment2D, self).__init__(sort, start, end)     

    def toString(self, printKey):
        s = ''
        if printKey and self.key is not None:
            s += '#'+str(self.key)+' '
        return s+'<'+str(self.start)+','+str(self.end)+'>'
                
    def getValue(self):
        return str(self)

    def printVal(self):
        return (str(self.start)+ ','+ str(self.end)) # + ', direction: ('+str(self.direction.x)+','+str(self.direction.y)+ '), unitDir: ('+str(self.unitDir.x)+','+str(self.unitDir.y)+ '), root: ('+str(self.root.x)+','+str(self.root.y)+ ')'

    def printRhino(self, dX = 0, dY = 0):
        return '_Line '+str(self.start.x + dX)+','+str(self.start.y + dY)+'  '+str(self.end.x + dX)+','+str(self.end.y + dY)

    def equals(self, other):
        return (self.start == other.start) and (self.end == other.end)
        
    def orientationDir(self):
        if self.reverse:
            return self.direction.inverse()
        return self.direction
        
    def isNil(self):
        if self.end is None or self.start is None:
            return True
        return self.end.x == self.start.x and self.end.y == self.start.y
        
    """
    compare the value of two line segments
    first compare using unit directions
    if unit directions are equal, compare using roots
    if roots are equal, compare the start points
    if starts points are equal, compare the end points
    *this function is named compareInterval and not compareValue since there
    is already a compareValue function in the line 
    """
    def compareInterval(self, other):
        c = self.unitDir.compareValue(other.unitDir) 
        if c != 0:
            return c
        c = self.root.compareValue(other.root) 
        if c != 0:
            return c
        c = self.start.compareValue(other.start) 
        if c != 0:
            return c
        c = self.end.compareValue(other.end) 
        if c != 0:
            return c
        return 0
        
    """
    check if the line segment contains a point
    """
    def containsPoint(self,point):
        if not point.colinear(self.start,self.end):
            return False
        if self.start.compareValue(point) < 1 and self.end.compareValue(point) > -1:
            return True
        return False
        
    def compareEnd(self, other):
        return self.end.compareValue(other.end) 
            
    """
    find the length of the line segment
    """
    def length(self):
        lengthVector = self.end.subtract(self.start)
        return lengthVector.dotproduct(lengthVector) ** c0p5
        
    """
    find the length of the line segment
    """
    def squareLength(self):
        lengthVector = self.end.subtract(self.start)
        return lengthVector.dotproduct(lengthVector)
        
    """
    check if two line segments touch one another
             o-----------o  
    o--------o
    
    or
        o-----o
              o------o
    """
    def touches(self,other):
        return self.compareValue(other) == 0 and (self.end == other.start or self.start == other.end) 
    
    """
    check if this line segment contains the other line segment
        o---------------------o
                o--------o
    or
        o---------------------o
        o---------o
    
    or
        o----------------------o
                o--------------o
    """
    def contains(self,other):
        return self.compareValue(other) == 0 and (self.end.compareValue(other.end) >= 0 and self.start.compareValue(other.start) <= 0) 
    
    """
    check if two line segments are disjoint
    o------------o
                        o----------o
    """
    def disjoint(self,other):
        return self.compareValue(other) == 0 and (self.end.compareValue(other.start) == -1 or self.start.compareValue(other.end) == 1)
    
    """
    check if two line segments are align
        o------------------o
            o--------------o    
    or
        o------------------o
        o---------o    
    or
                o----------o
        o------------------o 
    or 
        o-----o
        o------------------o
    """
    def align(self, other):
        return self.compareValue(other) == 0 and (self.end == other.end or self.start == other.start)
        
    """
    find the common of two line segment:
    o--------------o
            o-----------o
    -->     o------o        
    """
    def common(self,other):
        if self.disjoint(other) or self.touches(other):
            return False
        
        new = self.duplicateInd(self.sort)
        if self.end.compareValue(other.end) == 1:
            new.end = other.end
        if self.start.compareValue(other.start) == -1:
            new.start = other.start
            
        return new
        
        if self.end.compareValue(other.end) == 1:
            self.end = other.end
        if self.start.compareValue(other.start) == -1:
            self.start = other.start
        return True
    
    
    """
    combine two line segments into one (two lines cannot be disjoint)
    e.g.
            o----------o
                       o-------o      
       ---> o------------------o
    """
    def combine(self,other):
        if self.disjoint(other):
            return self
        nStart = self.start
        nEnd = self.end
        if self.end.compareValue(other.end) == -1:
            nEnd = other.end
        if self.start.compareValue(other.start) == 1:
            nStart = other.start
        return self.__class__(self.sort, nStart, nEnd)
        
    """
    find the complement of two lines
    e.g.
            o-------------o
                  o------------o      
       ---> o-----o       o----o
    or
            o------------------o
                    o----------o      
       ---> o-------o  
    or
            o------------------o
            o----------o      
       --->            o-------o  
    """
    def complement(self,other):
        result = []
        if self.start.compareValue(other.start) == -1 and self.end.compareValue(other.start) == 1 :
            new = self.duplicateInd(self.sort)
            new.end = other.start
            result.append(new)
        if self.start.compareValue(other.end) == -1 and self.end.compareValue(other.end) == 1 :
            new = self.duplicateInd(self.sort)
            new.start = other.end
            result.append(new)
        return result
                
        
            
    @staticmethod
    def ccw(A,B,C):
        return (C.y-A.y) * (B.x-A.x) > (B.y-A.y) * (C.x-A.x)
    
    # Return true if line segments AB and CD intersect
    def intersectSegment(self, other):
        if other is None or type(other) != type(self) :
            return False
        if self.direction.parallel(other.direction):
            return False
        if (self.start.colinear(other.start, other.end)
        or self.end.colinear(other.start, other.end)
        or other.start.colinear(self.start, self.end)
        or other.end.colinear(self.start, self.end)):
            return True
        A = self.start
        B = self.end
        C = other.start
        D = other.end
        return self.ccw(A,C,D) != self.ccw(B,C,D) and self.ccw(A,B,C) != self.ccw(A,B,D)
    
        
    """
    find the midpoint of a line segment
    """
    def midpoint(self):
        return (float(((self.start.x+self.end.x)/c2).value),float(((self.start.y+self.end.y)/c2).value))
        
    def replicateInd(self, sort = 0):
        if sort == 0:
            sort = self.sort
        return self.__class__(sort, self.start, self.end)    
        
    """
    return the line that contains this segment
    """
    def codescriptor(self):
        line = line2D(self.sort, self.start, self.end)
        line.assoc = self
        return line
        

    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            start = self.start
            end = self.end    
            newLine = self.duplicateInd()
        else:
            start = self.start.transform(affine)
            end = self.end.transform(affine)         
            newLine = self.__class__(self.sort, start, end)
        
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is not None:
            start = transf.apply(start)
            end = transf.apply(end)
            newLine = self.__class__(self.sort, start, end)
        return newLine
                
                
    def containsInd(self, ind):
        if not self.contains(ind):
            return False
        if self.sort.type == 'sort-attribute':
            if not self.hasAttribute():
                if ind.hasAttribute():
                    return False
                else:
                    return True
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        start = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        end = vector2D(x,y)
            
        ind = lineSegment2D(sort, start, end)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line segment')
            
        reader.newToken() #next token
        
        return ind

class lineSegment3D(line3D):
    """
    initialize a line segment
    initialize a line containing it first
    find the lesser and greater points between the 2 points passed in (start 
    and end) and store them as self.start and self.end respectively
    """
    type = 'ind-lineSegment3D'
    def __init__(self, sort, start = vector3D(0,0,0), end = vector3D(0,0,0), *args, **kwargs):
        if end.type == 'ind-point3D':
            end = end.value
        if start.type == 'ind-point3D':
            start = start.value
        if end.compareValue(start) == -1:
            self.start = end
            self.end = start
            self.reverse = True
        elif end.compareValue(start) == 1:
            self.end = end
            self.start = start
            self.reverse = False
        else:
            raise ValueError('start point and end point of the line segment are the same')
        super(lineSegment3D, self).__init__(sort, start, end)     

    def toString(self, printKey):
        s = ''
        if printKey and self.key is not None:
            s += '#'+str(self.key)+' '
        return s+'<'+str(self.start)+','+str(self.end)+'>'
                    
    def getValue(self):
        return str(self)

    def printVal(self):
        return str(self.start)+ ','+ str(self.end) # + ', direction: ('+str(self.direction.x)+','+str(self.direction.y)+ '), unitDir: ('+str(self.unitDir.x)+','+str(self.unitDir.y)+ '), root: ('+str(self.root.x)+','+str(self.root.y)+ ')'

    def printRhino(self, dX = 0, dY = 0):
        return '_Line '+str(self.start.x + dX)+','+str(self.start.y + dY)+','+str(self.start.z)+'  '+str(self.end.x + dX)+','+str(self.end.y + dY)+','+str(self.end.z)

    def equals(self, other):
        return (self.start == other.start) and (self.end == other.end)
        
    def orientationDir(self):
        if self.reverse:
            return self.direction.inverse()
        return self.direction
        
    def __ne__(self, other):
        return not self == other
 
        
    def isNil(self):
        if self.end is None or self.start is None:
            return True
        return self.end == self.start
        
    """
    compare the value of two line segments
    first compare using unit directions
    if unit directions are equal, compare using roots
    if roots are equal, compare the start points
    if starts points are equal, compare the end points
    *this function is named compareInterval and not compareValue since there
    is already a compareValue function in the line 
    """
    def compareInterval(self, other):
        c = self.unitDir.compareValue(other.unitDir) 
        if c != 0:
            return c
        c = self.root.compareValue(other.root) 
        if c != 0:
            return c
        c = self.start.compareValue(other.start) 
        if c != 0:
            return c
        c = self.end.compareValue(other.end) 
        if c != 0:
            return c
        return 0
        
    """
    check if the line segment contains a point
    """
    def containsPoint(self,point):
        if not point.colinear(self.start,self.end):
            return False
        if self.start.compareValue(point) < 1 and self.end.compareValue(point) > -1:
            return True
        return False
        
    def compareEnd(self, other):
        return self.end.compareValue(other.end) 
            
    """
    find the length of the line segment
    """
    def length(self):
        lengthVector = self.end.subtract(self.start)
        return lengthVector.dotproduct(lengthVector) ** c0p5
        
    """
    find the length of the line segment
    """
    def squareLength(self):
        lengthVector = self.end.subtract(self.start)
        return lengthVector.dotproduct(lengthVector)
        
    """
    check if two line segments touch one another
             o-----------o  
    o--------o
    
    or
        o-----o
              o------o
    """
    def touches(self,other):
        return self.compareValue(other) == 0 and (self.end == other.start or self.start == other.end) 
    
    """
    check if this line segment contains the other line segment
        o---------------------o
                o--------o
    or
        o---------------------o
        o---------o
    
    or
        o----------------------o
                o--------------o
    """
    def contains(self,other):
        return self.compareValue(other) == 0 and (self.end.compareValue(other.end) >= 0 and self.start.compareValue(other.start) <= 0) 
    
    """
    check if two line segments are disjoint
    o------------o
                        o----------o
    """
    def disjoint(self,other):
        return self.compareValue(other) == 0 and (self.end.compareValue(other.start) == -1 or self.start.compareValue(other.end) == 1)
    
    """
    check if two line segments are align
        o------------------o
            o--------------o    
    or
        o------------------o
        o---------o    
    or
                o----------o
        o------------------o 
    or 
        o-----o
        o------------------o
    """
    def align(self, other):
        return self.compareValue(other) == 0 and (self.end == other.end or self.start == other.start)
        
    """
    find the common of two line segment:
    o--------------o
            o-----------o
    -->     o------o        
    """

    def common(self,other):
        if self.disjoint(other) or self.touches(other):
            return False
        
        new = self.duplicateInd(self.sort)
        if self.end.compareValue(other.end) == 1:
            new.end = other.end
        if self.start.compareValue(other.start) == -1:
            new.start = other.start
            
        return new
            
        if self.end.compareValue(other.end) == 1:
            self.end = other.end
        if self.start.compareValue(other.start) == -1:
            self.start = other.start
        return True
        
    """
    combine two line segments into one (two lines cannot be disjoint)
    e.g.
            o----------o
                       o-------o      
       ---> o------------------o
    """
    def combine(self,other):
        if self.disjoint(other):
            return self
        nStart = self.start
        nEnd = self.end
        if self.end.compareValue(other.end) == -1:
            nEnd = other.end
        if self.start.compareValue(other.start) == 1:
            nStart = other.start
        return self.__class__(self.sort, nStart, nEnd)
        
    """
    find the complement of two lines
    e.g.
            o-------------o
                  o------------o      
       ---> o-----o       o----o
    or
            o------------------o
                    o----------o      
       ---> o-------o  
    or
            o------------------o
            o----------o      
       --->            o-------o  
    """
    def complement(self,other):
        result = []
        if self.start.compareValue(other.start) == -1 and self.end.compareValue(other.start) == 1 :
            new = self.duplicateInd(self.sort)
            new.end = other.start
            result.append(new)
        if self.start.compareValue(other.end) == -1 and self.end.compareValue(other.end) == 1 :
            new = self.duplicateInd(self.sort)
            new.start = other.end
            result.append(new)
        return result
        
    def midpoint(self):
        return (float(((self.start.x+self.end.x)/c2).value),float(((self.start.y+self.end.y)/c2).value),float(((self.start.z+self.end.z)/c2).value))
            
    def replicateInd(self, sort = 0):
        if sort == 0:
            sort = self.sort
        return self.__class__(sort, self.start, self.end)    
        
    """
    return the line that contains this segment
    """
    def codescriptor(self):
        line = line3D(self.sort, self.start, self.end)
        line.assoc = self
        return line
        

    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            start = self.start
            end = self.end    
            newLine = self.duplicateInd()
        else:
            start = self.start.transform(affine)
            end = self.end.transform(affine)         
            newLine = self.__class__(self.sort, start, end)
        
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is not None:
            line = transf.apply(newLine)
            if line is not None:
                return line
            start = transf.apply(start)
            end = transf.apply(end)
            newLine = self.__class__(self.sort, start, end)
        return newLine

        
    def containsInd(self, ind):
        if not self.contains(ind):
            return False
        if self.sort.type == 'sort-attribute':
            if not self.hasAttribute():
                if ind.hasAttribute():
                    return False
                else:
                    return True
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y,z = vector3D.readVector(reader)
        start = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        end = vector3D(x,y,z)
            
        ind = lineSegment3D(sort, start, end)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line segment')
            
        reader.newToken() #next token
        
        return ind
